@@ -4,7 +4,7 @@ package LWPx::UserAgent::Cached;
 
 use Modern::Perl '2011';    ## no critic (Modules::ProhibitUseQuotedVersion)
 
-our $VERSION = '0.002';     # TRIAL VERSION
+our $VERSION = '0.005';     # TRIAL VERSION
 use utf8;
 
 ## no critic (Bangs::ProhibitCommentedOutCode)
@@ -28,6 +28,8 @@ use utf8;
 #pod inheriting from L<WWW::Mechanize|WWW::Mechanize>;
 #pod instead it is just a direct subclass of
 #pod L<LWP::UserAgent|LWP::UserAgent>.
+#pod As of version 0.005 it has limited support for HTTP/1.1
+#pod C<ETag>/C<If-None-Match> cache control headers.
 #pod
 #pod =head1 SEE ALSO
 #pod
@@ -46,14 +48,15 @@ use utf8;
 #pod =cut
 
 use CHI;
-use HTTP::Status qw(HTTP_OK HTTP_MOVED_PERMANENTLY);
+use HTTP::Status qw(HTTP_OK HTTP_MOVED_PERMANENTLY HTTP_NOT_MODIFIED);
 
 # work around RT#43310
 ## no critic (Subroutines::ProhibitCallsToUndeclaredSubs)
 use List::Util 1.33 'any';
-use Storable qw(nfreeze thaw);
+use Sereal qw(sereal_encode_with_object sereal_decode_with_object);
 use Moo 1.004005;
-use Types::Standard qw(Bool HasMethods HashRef Maybe);
+use Sub::Quote 1.005000 'qsub';
+use Types::Standard qw(Bool HasMethods HashRef InstanceOf Maybe);
 use namespace::clean;
 extends 'LWP::UserAgent';
 
@@ -140,12 +143,20 @@ has ref_in_cache_key => ( is => 'rw', isa => Bool, default => 0 );
 
 has positive_cache => ( is => 'rw', isa => Bool, default => 1 );
 
+for (qw(Encoder Decoder)) {
+    has "_\l$_" => (
+        is      => 'lazy',
+        isa     => InstanceOf ["Sereal::$_"],
+        default => qsub "Sereal::$_->new",
+    );
+}
+
 #pod =head1 HANDLERS
 #pod
 #pod This module works by adding C<request_send> and C<response_done>
-#pod L<handlers|LWP::UserAgent/Handlers> method that run on successful HTTP
-#pod C<GET> requests. If you need to modify or remove these handlers you may use
-#pod L<LWP::UserAgent's C<handlers>|LWP::UserAgent/Handlers> method.
+#pod L<handlers|LWP::UserAgent/Handlers> method that run on successful
+#pod HTTP C<GET> requests. If you need to modify or remove these handlers you may
+#pod use L<LWP::UserAgent's C<handlers>|LWP::UserAgent/Handlers> method.
 #pod
 #pod =for Pod::Coverage BUILD
 #pod
@@ -154,63 +165,102 @@ has positive_cache => ( is => 'rw', isa => Bool, default => 1 );
 sub BUILD {
     my $self = shift;
 
-    $self->add_handler(    # load from cache on each GET request
-        request_send => sub {
-            my $original_request = shift;
-            my $request          = $original_request;
-
-            $self->_set_is_cached(0);
-
-            if ( not $self->ref_in_cache_key ) {
-                my $clone = $request->clone;
-                $clone->header( Referer => undef );
-                $request = $clone->as_string;
-            }
-
-            my $response = $self->cache->get($request);
-            if ($response) { $response = thaw($response) }
-            return
-                   if not($response)
-                or $response->code < HTTP_OK
-                or $response->code > HTTP_MOVED_PERMANENTLY;
-
-            $self->_set_is_cached(1);
-            return $response;
-        },
+    $self->add_handler(
+        request_send => $self->_closure_get_cache,
         ( m_method => 'GET' ),
     );
+    $self->add_handler(
+        response_done => $self->_closure_set_cache,
+        ( m_method => 'GET', m_code => 2 ),
+    );
+    $self->add_handler(
+        response_done => $self->_closure_not_modified,
+        ( m_code => HTTP_NOT_MODIFIED ),
+    );
 
-    $self->add_handler(    # save to cache after successful GET
-        response_done => sub {
-            return if not my $response = shift;
+    return;
+}
 
-            if ( not $response->header('client-transfer-encoding')
+# load from cache on each GET request
+sub _closure_get_cache {
+    my $self = shift;
+    return sub {
+        my ($request) = @_;
+        $self->_set_is_cached(0);
+        if ( not $self->ref_in_cache_key ) {
+            my $clone = $request->clone;
+            $clone->header( Referer => undef );
+            $request = $clone->as_string;
+        }
+
+        my $response = $self->cache->get("$request");
+        $response &&= sereal_decode_with_object( $self->_decoder, $response );
+        return if not $response;
+
+        my $etag;
+        if ( $etag = $response->header('etag') ) {
+            $_[0]->header( if_none_match => $etag );
+        }
+
+        return
+               if $etag
+            or $response->code < HTTP_OK
+            or $response->code > HTTP_MOVED_PERMANENTLY;
+        $self->_set_is_cached(1);
+        return $response;
+    };
+}
+
+# save to cache after successful GET
+sub _closure_set_cache {
+    my $self = shift;
+    return sub {
+        return if not my $response = shift;
+
+        if (not( $response->header('client-transfer-encoding')
                 and 'ARRAY' eq
                 ref $response->header('client-transfer-encoding')
                 and any { 'chunked' eq $_ }
                 @{ $response->header('client-transfer-encoding') } )
-            {
-                for ( $response->header('size') ) {
-                    return
-                        if not defined and $self->cache_undef_content_length;
-                    return
-                        if 0 == $_
-                        and not $self->cache_zero_content_length;
-                    return
-                        if $_ != length $response->content
-                        and not $self->cache_mismatch_content_length;
-                }
+            )
+        {
+            for ( $response->header('size') ) {
+                return
+                    if not defined and $self->cache_undef_content_length;
+                return
+                    if 0 == $_
+                    and not $self->cache_zero_content_length;
+                return
+                    if $_ != length $response->content
+                    and not $self->cache_mismatch_content_length;
             }
+        }
 
-            $response->decode;
-            $self->cache->set( $response->request->as_string,
-                nfreeze($response) );
-            return;
-        },
-        ( m_method => 'GET', m_code => 2 ),
-    );
+        $response->decode;
+        $self->cache->set( $response->request->as_string =>
+                sereal_encode_with_object( $self->_encoder, $response ) );
+        return;
+    };
+}
 
-    return;
+# handle HTTP 304 response, possibly from cache
+sub _closure_not_modified {
+    my $self = shift;
+    return sub {
+        my $request = $_[0]->request;
+        $self->_set_is_cached(0);
+
+        $request->header( if_none_match => undef );
+        my $response = $self->cache->get( $request->as_string );
+        $response &&= sereal_decode_with_object( $self->_decoder, $response );
+        if ( not $response ) {
+            $request->header( if_none_match => undef );
+            $response = $self->request($request);
+        }
+
+        $self->_set_is_cached(1);
+        $_[0] = $response;
+    };
 }
 
 #pod =for Pod::Coverage FOREIGNBUILDARGS
@@ -240,7 +290,7 @@ LWPx::UserAgent::Cached - Subclass of LWP::UserAgent that caches HTTP GET reques
 
 =head1 VERSION
 
-version 0.002
+version 0.005
 
 =head1 SYNOPSIS
 
@@ -261,6 +311,8 @@ L<WWW::Mechanize::Cached|WWW::Mechanize::Cached> but without
 inheriting from L<WWW::Mechanize|WWW::Mechanize>;
 instead it is just a direct subclass of
 L<LWP::UserAgent|LWP::UserAgent>.
+As of version 0.005 it has limited support for HTTP/1.1
+C<ETag>/C<If-None-Match> cache control headers.
 
 =head1 ATTRIBUTES
 
@@ -322,9 +374,9 @@ Inspiration for this class.
 =head1 HANDLERS
 
 This module works by adding C<request_send> and C<response_done>
-L<handlers|LWP::UserAgent/Handlers> method that run on successful HTTP
-C<GET> requests. If you need to modify or remove these handlers you may use
-L<LWP::UserAgent's C<handlers>|LWP::UserAgent/Handlers> method.
+L<handlers|LWP::UserAgent/Handlers> method that run on successful
+HTTP C<GET> requests. If you need to modify or remove these handlers you may
+use L<LWP::UserAgent's C<handlers>|LWP::UserAgent/Handlers> method.
 
 =for Pod::Coverage BUILD
 
